/*
 * QEMU backup
 *
 * Copyright (C) 2013 Proxmox Server Solutions
 *
 * Authors:
 *  Dietmar Maurer (dietmar@proxmox.com)
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>

#include "trace.h"
#include "block/block.h"
#include "block/block_int.h"
#include "block/blockjob.h"
#include "qemu/ratelimit.h"
#include "qemu/bitmap.h"
#include "qapi-event.h"

#include "block/coroutine_int.h"

#define BACKUP_CLUSTER_BITS 16
#define BACKUP_CLUSTER_SIZE (1 << BACKUP_CLUSTER_BITS)
#define BACKUP_SECTORS_PER_CLUSTER (BACKUP_CLUSTER_SIZE / BDRV_SECTOR_SIZE)
#define BACKUP_MAX_QUERY_SECTORS \
    ((INT_MAX & ~(BACKUP_CLUSTER_SIZE - 1)) >> BDRV_SECTOR_BITS)

#define SLICE_TIME 100000000ULL /* ns */
#define BACKUP_LOOP_DELAY 100000000ULL /* ns */
#define WRITE_NOTIFY_TIMEOUT_NS 30000000000LL

#define NB_WORKERS 24

typedef struct NotifierRequest {
    int64_t start; /* in clusters */
    int64_t end; /* in clusters */
    int nb_wait; /* awaited clusters */
    Coroutine *notif;
    QSIMPLEQ_ENTRY(NotifierRequest) list;
} NotifierRequest;

typedef struct BackupBlockJob {
    BlockJob common;
    BlockDriverState *target;
    /* bitmap for sync=incremental */
    BdrvDirtyBitmap *sync_bitmap;
    MirrorSyncMode sync_mode;
    RateLimit limit;
    BlockdevOnError on_source_error;
    BlockdevOnError on_target_error;
    uint64_t sectors_read;
    bool compress;

    HBitmap *copy_bitmap;
    HBitmapIter linear_hbi;

    int async_status;
    CoQueue free_workers;
    bool delayed;
    int running_workers;

    unsigned long *notif_wait_bitmap;
    QSIMPLEQ_HEAD(, NotifierRequest) notifier_reqs;
    NotifierRequest *current_notif;
    HBitmapIter current_notif_hbi;
} BackupBlockJob;

static inline void set_notif_cur(BackupBlockJob *job, NotifierRequest *nr)
{
    if (nr != NULL) {
        hbitmap_iter_init(&job->current_notif_hbi, job->copy_bitmap, nr->start);
    }
    job->current_notif = nr;
}

/* NULL result means that request is already done */
static NotifierRequest *add_notif_req(BackupBlockJob *job,
                                      int64_t start, int64_t end,
                                      Coroutine *notif)
{
    NotifierRequest *nr;
    int nb_wait = bitmap_count_between(job->notif_wait_bitmap, start, end) +
                  hbitmap_count_between(job->copy_bitmap, start, end);

    if (nb_wait == 0) {
        return NULL;
    }

    nr = g_new0(NotifierRequest, 1);

    nr->start = start;
    nr->end = end;
    nr->notif = notif;
    nr->nb_wait = nb_wait;

    QSIMPLEQ_INSERT_TAIL(&job->notifier_reqs, nr, list);
    if (job->current_notif == NULL) {
        set_notif_cur(job, nr);
    }

    return nr;
}

static void detach_notif_req(BackupBlockJob *job, NotifierRequest *req)
{
    assert(req);

    if (job->current_notif == req) {
        set_notif_cur(job, QSIMPLEQ_NEXT(req, list));
    }

    QSIMPLEQ_REMOVE(&job->notifier_reqs, req, NotifierRequest, list);
}

static int64_t next_notif_cluster(BackupBlockJob *job)
{
    while (job->current_notif != NULL) {
        int64_t cluster = hbitmap_iter_next(&job->current_notif_hbi);
        if (cluster != -1 && cluster < job->current_notif->end) {
            return cluster;
        }

        /* nothing to do for current notifier */
        set_notif_cur(job, QSIMPLEQ_NEXT(job->current_notif, list));
    }

    /* nothing to do for notifiers */
    return -1;
}

static void finish_cluster_for_notif(BackupBlockJob *job, int64_t cluster)
{
    NotifierRequest *req, *next;
    QSIMPLEQ_HEAD(, NotifierRequest) finished_reqs;
    QSIMPLEQ_INIT(&finished_reqs);

    trace_backup_finish_cluster_for_notif(job, cluster);

    clear_bit(cluster, job->notif_wait_bitmap);

    /* handle notifiers */
    QSIMPLEQ_FOREACH_SAFE(req, &job->notifier_reqs, list, next) {
        if (cluster >= req->start && cluster < req->end) {
            assert(req->nb_wait > 0);
            req->nb_wait--;
            if (req->nb_wait == 0) {
                detach_notif_req(job, req);
                QSIMPLEQ_INSERT_TAIL(&finished_reqs, req, list);
            }
        }
    }

    while ((req = QSIMPLEQ_FIRST(&finished_reqs))) {
        QSIMPLEQ_REMOVE(&finished_reqs, req, NotifierRequest, list);
        qemu_coroutine_enter(req->notif, NULL);
    }
}

static void release_pending_notifiers(BackupBlockJob *job)
{
    NotifierRequest *req, *next;
    QSIMPLEQ_HEAD(, NotifierRequest) finished_reqs;
    QSIMPLEQ_INIT(&finished_reqs);

    QSIMPLEQ_FOREACH_SAFE(req, &job->notifier_reqs, list, next) {
        req->nb_wait = 0;
        detach_notif_req(job, req);
        QSIMPLEQ_INSERT_TAIL(&finished_reqs, req, list);
    }

    while ((req = QSIMPLEQ_FIRST(&finished_reqs))) {
        QSIMPLEQ_REMOVE(&finished_reqs, req, NotifierRequest, list);
        qemu_coroutine_enter(req->notif, NULL);
    }
}


static void backup_report_error(BackupBlockJob *job, bool is_read)
{
    /* Taken from block_job_error_action.
     * Actually, this call is equal to call block_job_error_action, because
     * the action should always be 'report' for backup. But is is safer to send
     * event directly. */
    qapi_event_send_block_job_error(bdrv_get_device_name(job->common.bs),
            is_read ? IO_OPERATION_TYPE_READ : IO_OPERATION_TYPE_WRITE,
            BLOCK_ERROR_ACTION_REPORT, &error_abort);
}

static int coroutine_fn backup_do_read(BackupBlockJob *job,
                                       int64_t sector_num, int nb_sectors,
                                       bool is_write_notifier,
                                       QEMUIOVector *qiov)
{
    BlockDriverState *bs = job->common.bs;
    int ret;

    if (is_write_notifier) {
        ret = bdrv_co_readv_no_serialising(bs, sector_num, nb_sectors, qiov);
    } else {
        ret = bdrv_co_readv(bs, sector_num, nb_sectors, qiov);
    }
    if (ret < 0) {
        trace_backup_do_read_fail(job,
                sector_num / BACKUP_SECTORS_PER_CLUSTER, ret);

        backup_report_error(job, true);
    }

    return ret;
}

static int coroutine_fn backup_do_write(BackupBlockJob *job,
                                        int64_t sector_num, int nb_sectors,
                                        QEMUIOVector *qiov)
{
    int ret;
    assert(qiov->niov == 1);

    if (buffer_is_zero(qiov->iov->iov_base, qiov->iov->iov_len)) {
        ret = bdrv_co_write_zeroes(job->target, sector_num, nb_sectors,
                                   BDRV_REQ_MAY_UNMAP);
    } else if (job->compress) {
        ret = bdrv_co_write_compressed(job->target, sector_num, nb_sectors,
                                       qiov);
    } else {
        ret = bdrv_co_writev(job->target, sector_num, nb_sectors, qiov);
    }
    if (ret < 0) {
        trace_backup_do_write_fail(job,
                sector_num / BACKUP_SECTORS_PER_CLUSTER, ret);

        backup_report_error(job, false);
    }

    return ret;
}

static void coroutine_fn backup_copy_cluster(BlockDriverState *bs,
        int64_t cluster, bool is_write_notifier)
{
    BackupBlockJob *job = (BackupBlockJob *)bs->job;
    int ret;
    int n;
    struct iovec iov;
    QEMUIOVector bounce_qiov;
    void *bounce_buffer = qemu_blockalign(bs, BACKUP_CLUSTER_SIZE);

    trace_backup_copy_cluster_enter(job, cluster);

    n = MIN(BACKUP_SECTORS_PER_CLUSTER,
            job->common.len / BDRV_SECTOR_SIZE -
            cluster * BACKUP_SECTORS_PER_CLUSTER);

    iov.iov_base = bounce_buffer;
    iov.iov_len = n * BDRV_SECTOR_SIZE;
    qemu_iovec_init_external(&bounce_qiov, &iov, 1);

    ret = backup_do_read(job, cluster * BACKUP_SECTORS_PER_CLUSTER, n,
                         is_write_notifier, &bounce_qiov);
    if (ret < 0) {
        goto out;
    }

    if (job->sync_mode != MIRROR_SYNC_MODE_NONE) {
        finish_cluster_for_notif(job, cluster);
    }

    ret = backup_do_write(job, cluster * BACKUP_SECTORS_PER_CLUSTER, n,
                          &bounce_qiov);
    if (ret < 0) {
        goto out;
    }

    if (job->sync_mode == MIRROR_SYNC_MODE_NONE) {
        finish_cluster_for_notif(job, cluster);
    }

    /* Publish progress, guest I/O counts as progress too.  Note that the
     * offset field is an opaque progress value, it is not a disk offset.
     */
    job->sectors_read += n;
    job->common.offset += n * BDRV_SECTOR_SIZE;
    ret = 0;

out:
    if (ret < 0 && job->async_status == 0) {
        job->async_status = ret;
    }
    qemu_vfree(bounce_buffer);

    trace_backup_copy_cluster_return(job, cluster, ret);
}

static int coroutine_fn backup_before_write_notify(
        NotifierWithReturn *notifier,
        void *opaque)
{
    BdrvTrackedRequest *tr = opaque;
    NotifierRequest *nr;
    BackupBlockJob *job = (BackupBlockJob *)tr->bs->job;
    int64_t start = tr->offset >> BACKUP_CLUSTER_BITS;
    int64_t end = DIV_ROUND_UP(tr->offset + tr->bytes, BACKUP_CLUSTER_SIZE);

    assert((tr->offset & (BDRV_SECTOR_SIZE - 1)) == 0);
    assert((tr->bytes & (BDRV_SECTOR_SIZE - 1)) == 0);

    nr = add_notif_req(job, start, end, qemu_coroutine_self());

    if (nr == NULL) {
        trace_backup_before_write_notify_skip(job, tr->offset, tr->bytes);
    } else {
        trace_backup_before_write_notify_start(job, tr->offset, tr->bytes,
                                               nr, nr->start, nr->end, nr->nb_wait);

        qemu_co_queue_restart_all(&job->free_workers);
        co_aio_sleep_ns(bdrv_get_aio_context(job->common.bs),
                        QEMU_CLOCK_REALTIME, WRITE_NOTIFY_TIMEOUT_NS);
        if (nr->nb_wait > 0) {
            /* timer expired and read request not finished */
            if (job->async_status == 0) {
                job->async_status = -ETIME;
            }
            detach_notif_req(job, nr);
            trace_backup_before_write_notify_timeout(job, nr, nr->start, nr->end, nr->nb_wait);
        } else {
            trace_backup_before_write_notify_success(job, nr, nr->start, nr->end, nr->nb_wait);
        }
        g_free(nr);
    }

    return 0;
}

static void backup_set_speed(BlockJob *job, int64_t speed, Error **errp)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);

    if (speed < 0) {
        error_set(errp, QERR_INVALID_PARAMETER, "speed");
        return;
    }
    ratelimit_set_speed(&s->limit, speed / BDRV_SECTOR_SIZE, SLICE_TIME);
}

static void backup_iostatus_reset(BlockJob *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);

    bdrv_iostatus_reset(s->target);
}

static void backup_cleanup_sync_bitmap(BackupBlockJob *job, int ret)
{
    BdrvDirtyBitmap *bm;
    BlockDriverState *bs = job->common.bs;

    if (ret < 0 || block_job_is_cancelled(&job->common)) {
        /* Merge the successor back into the parent, delete nothing. */
        bm = bdrv_reclaim_dirty_bitmap(bs, job->sync_bitmap, NULL);
        assert(bm);
    } else {
        /* Everything is fine, delete this bitmap and install the backup. */
        bm = bdrv_dirty_bitmap_abdicate(bs, job->sync_bitmap, NULL);
        assert(bm);
    }
}

static void backup_commit(BlockJob *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);
    if (s->sync_bitmap) {
        backup_cleanup_sync_bitmap(s, 0);
    }
}

static void backup_abort(BlockJob *job)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);
    if (s->sync_bitmap) {
        backup_cleanup_sync_bitmap(s, -1);
    }
}

static const BlockJobDriver backup_job_driver = {
    .instance_size  = sizeof(BackupBlockJob),
    .job_type       = BLOCK_JOB_TYPE_BACKUP,
    .set_speed      = backup_set_speed,
    .iostatus_reset = backup_iostatus_reset,
    .commit         = backup_commit,
    .abort          = backup_abort,
};

typedef struct {
    int ret;
} BackupCompleteData;

static void backup_complete(BlockJob *job, void *opaque)
{
    BackupBlockJob *s = container_of(job, BackupBlockJob, common);
    BackupCompleteData *data = opaque;

    bdrv_unref(s->target);

    block_job_completed(job, data->ret);
    g_free(data);
}

static void backup_sleep(BackupBlockJob *job)
{
    /* we need to yield so that bdrv_drain_all() returns.
     * (without, VM does not reboot)
     */
    if (job->common.speed) {
        uint64_t delay_ns = ratelimit_calculate_delay(&job->limit,
                                                      job->sectors_read);
        job->sectors_read = 0;
        job->delayed = true;
        block_job_sleep_ns(&job->common, QEMU_CLOCK_REALTIME, delay_ns);
        job->delayed = false;
    } else {
        block_job_sleep_ns(&job->common, QEMU_CLOCK_REALTIME, 0);
    }
}

static bool coroutine_fn yield_and_check(BackupBlockJob *job)
{
    if (block_job_is_cancelled(&job->common)) {
        return true;
    }

    backup_sleep(job);

    if (block_job_is_cancelled(&job->common)) {
        return true;
    }

    return false;
}

/* init copy_bitmap as sync_bitmap */
static void backup_incremental_init_copy_bitmap(BackupBlockJob *job)
{
    int64_t sector;
    HBitmapIter hbi;
    uint32_t sect_gran =
        bdrv_dirty_bitmap_granularity(job->sync_bitmap) >> BDRV_SECTOR_BITS;
    int64_t sz = bdrv_dirty_bitmap_size(job->sync_bitmap);
    uint32_t cl_gran = MAX(1, sect_gran / BACKUP_SECTORS_PER_CLUSTER);

    bdrv_dirty_iter_init(job->sync_bitmap, &hbi);
    while ((sector = hbitmap_iter_next(&hbi)) != -1) {
        int64_t cluster = sector / BACKUP_SECTORS_PER_CLUSTER;
        int64_t next_sector = (cluster + cl_gran) * BACKUP_SECTORS_PER_CLUSTER;

        hbitmap_set(job->copy_bitmap, cluster, cl_gran);

        if (next_sector >= sz) {
            break;
        }

        bdrv_set_dirty_iter(&hbi,
                            (cluster + cl_gran) * BACKUP_SECTORS_PER_CLUSTER);
    }

    job->common.offset = job->common.len -
                         hbitmap_count(job->copy_bitmap) * BACKUP_CLUSTER_SIZE;
}

static void backup_skip_clusters(BackupBlockJob *job,
                                 int64_t start, int64_t end)
{
    int n;

    if (end * BACKUP_SECTORS_PER_CLUSTER > job->common.len) {
        end--;
        n = job->common.len - end * BACKUP_SECTORS_PER_CLUSTER;
        assert(n > 0);

        if (hbitmap_get(job->copy_bitmap, end)) {
            trace_backup_skip_cluster(job, end);
            hbitmap_reset(job->copy_bitmap, end, 1);
            job->common.offset += n * BDRV_SECTOR_SIZE;
            finish_cluster_for_notif(job, end);
        }
    }

    n = BACKUP_SECTORS_PER_CLUSTER;
    for ( ; start < end; start++) {
        if (!hbitmap_get(job->copy_bitmap, start)) {
            continue;
        }

        trace_backup_skip_cluster(job, start);
        hbitmap_reset(job->copy_bitmap, start, 1);
        job->common.offset += n * BDRV_SECTOR_SIZE;
        finish_cluster_for_notif(job, start);
    }
}

static void backup_skip_unallocated_clusters(BackupBlockJob *job,
                                             BlockDriverState *base)
{
    BlockDriverState *bs = job->common.bs;
    int64_t end = job->common.len >> BDRV_SECTOR_BITS;
    int64_t sector = 0;

    while (sector < end) {
        int ret;
        int n;
        int nb_sectors = MIN(BACKUP_MAX_QUERY_SECTORS,
                             end - sector);
        int64_t cluster = sector / BACKUP_SECTORS_PER_CLUSTER;

        if (!hbitmap_get(job->copy_bitmap, cluster)) {
            sector += BACKUP_SECTORS_PER_CLUSTER;
            continue;
        }

        if (yield_and_check(job) || job->async_status < 0) {
            return;
        }

        ret = bdrv_is_allocated_above(bs, base, sector, nb_sectors, &n);
        if (job->async_status < 0) {
            return;
        }
        if (ret < 0) {
            job->async_status = ret;
            return;
        }

        if (ret == 1 || n < BACKUP_SECTORS_PER_CLUSTER) {
            sector += DIV_ROUND_UP(n, BACKUP_SECTORS_PER_CLUSTER) *
                          BACKUP_SECTORS_PER_CLUSTER;
            continue;
        }

        if (nb_sectors + n < job->common.len) {
            n -= n % BACKUP_SECTORS_PER_CLUSTER;
        }

        backup_skip_clusters(job, cluster,
                DIV_ROUND_UP(sector + n, BACKUP_SECTORS_PER_CLUSTER));

        assert(n > 0);
        sector += n;
    }
}

#define BACKUP_WORKER_STOP -1
#define BACKUP_WORKER_PAUSE -2

static inline int64_t backup_get_work(BackupBlockJob *job,
                                      bool *is_write_notifier)
{
    int64_t cluster;

    assert(is_write_notifier != NULL);

    if (block_job_is_cancelled(&job->common) || job->async_status != 0) {
        return BACKUP_WORKER_STOP;
    }

    cluster = next_notif_cluster(job);
    *is_write_notifier = cluster != -1;
    if (cluster == -1) {
        if (block_job_is_paused(&job->common) || job->delayed ||
            job->sync_mode == MIRROR_SYNC_MODE_NONE) {
            return BACKUP_WORKER_PAUSE;
        }

        cluster = hbitmap_iter_next(&job->linear_hbi);
        if (cluster == -1) {
            return BACKUP_WORKER_STOP;
        }
    }

    assert(hbitmap_get(job->copy_bitmap, cluster));
    hbitmap_reset(job->copy_bitmap, cluster, 1);
    set_bit(cluster, job->notif_wait_bitmap);

    return cluster;
}

static void backup_worker_pause(BackupBlockJob *job)
{
    qemu_co_queue_wait(&job->free_workers);
}

static void coroutine_fn backup_worker_co(void *opaque)
{
    BackupBlockJob *job = opaque;

    job->running_workers++;

    while (true) {
        bool is_write_notifier;
        int64_t cluster = backup_get_work(job, &is_write_notifier);

        trace_backup_worker_co(job, cluster, is_write_notifier);

        switch (cluster) {
        case BACKUP_WORKER_STOP:
            job->running_workers--;
            return;
        case BACKUP_WORKER_PAUSE:
            backup_worker_pause(job);
            break;
        default:
            backup_copy_cluster(job->common.bs, cluster, is_write_notifier);
        }
    }
}

/* this loop runs in block job coroutine (job->common.co) and handles
 * block-job specific pauses, delays and yields. It doesn't do real data copy.
 */
static void coroutine_fn backup_loop(BackupBlockJob *job)
{
    while (job->running_workers > 0) {
        backup_sleep(job);

        /* we may be resumed after pause or delay, so we should resume paused
         * workers */
        qemu_co_queue_restart_all(&job->free_workers);

        co_aio_sleep_ns(bdrv_get_aio_context(job->common.bs),
                        QEMU_CLOCK_REALTIME, BACKUP_LOOP_DELAY);
    }
}

static void coroutine_fn backup_run(void *opaque)
{
    BackupBlockJob *job = opaque;
    BackupCompleteData *data;
    BlockDriverState *bs = job->common.bs;
    BlockDriverState *target = job->target;
    BlockdevOnError on_target_error = job->on_target_error;
    NotifierWithReturn before_write = {
        .notify = backup_before_write_notify,
    };
    int64_t end;
    bool is_top = job->sync_mode == MIRROR_SYNC_MODE_TOP;
    bool is_full = job->sync_mode == MIRROR_SYNC_MODE_FULL;
    int i;

    qemu_co_queue_init(&job->free_workers);
    QSIMPLEQ_INIT(&job->notifier_reqs);

    end = DIV_ROUND_UP(job->common.len, BACKUP_CLUSTER_SIZE);

    job->copy_bitmap = hbitmap_alloc(end, 0);
    job->notif_wait_bitmap = bitmap_new(end);

    bdrv_set_enable_write_cache(target, true);
    bdrv_set_on_error(target, on_target_error, on_target_error);
    bdrv_iostatus_enable(target);

    bdrv_add_before_write_notifier(bs, &before_write);

    /* init copy_bitmap */
    if (job->sync_mode == MIRROR_SYNC_MODE_INCREMENTAL) {
        backup_incremental_init_copy_bitmap(job);
    } else {
        hbitmap_set(job->copy_bitmap, 0, end);
    }
    hbitmap_iter_init(&job->linear_hbi, job->copy_bitmap, 0);

    for (i = 0; i < NB_WORKERS; ++i) {
        Coroutine *co = qemu_coroutine_create(backup_worker_co);
        qemu_coroutine_enter(co, job);
    }

    if (job->sync_mode == MIRROR_SYNC_MODE_NONE) {
        while (!block_job_is_cancelled(&job->common)) {
            /* Yield until the job is cancelled.  We just let backup workers
             * service CoW requests. */
            job->common.busy = false;
            qemu_coroutine_yield();
            job->common.busy = true;
        }
    } else if (is_top || (is_full && bdrv_has_zero_init(target))) {
        BlockDriverState *base = is_top ? bs->backing_hd : NULL;
        backup_skip_unallocated_clusters(job, base);
    }

    /* wait for all workers to exit */
    backup_loop(job);

    notifier_with_return_remove(&before_write);

    /* in case of error or cancel there may be pending notifiers */
    release_pending_notifiers(job);

    hbitmap_free(job->copy_bitmap);
    g_free(job->notif_wait_bitmap);

    if (job->async_status == 0) {
        job->common.success = true;
        while (!block_job_txn_all_success(&job->common)) {
            if (block_job_is_cancelled(&job->common)) {
                break;
            }

            block_job_sleep_ns(&job->common, QEMU_CLOCK_REALTIME, SLICE_TIME);
        }
    }

    bdrv_iostatus_disable(target);
    bdrv_op_unblock_all(target, job->common.blocker);

    data = g_malloc(sizeof(*data));
    data->ret = job->async_status;
    block_job_defer_to_main_loop(&job->common, backup_complete, data);
}

void backup_start(BlockDriverState *bs, BlockDriverState *target,
                  int64_t speed, MirrorSyncMode sync_mode,
                  BdrvDirtyBitmap *sync_bitmap,
                  bool compress,
                  BlockdevOnError on_source_error,
                  BlockdevOnError on_target_error,
                  BlockCompletionFunc *cb, void *opaque,
                  BlockJobTxn *txn, Error **errp)
{
    int64_t len;

    assert(bs);
    assert(target);
    assert(cb);

    if (bs == target) {
        error_setg(errp, "Source and target cannot be the same");
        return;
    }

    /* virtuozzo: to simplify async backup scheme we have dropped
     * retrying of io operations in case of errors.
     */
    if (on_source_error != BLOCKDEV_ON_ERROR_REPORT ||
        on_target_error != BLOCKDEV_ON_ERROR_REPORT) {
        error_setg(errp, "Only 'report' value of on-source-error and "
                         "on-target-error parameters is supported.");
        return;
    }

    if ((on_source_error == BLOCKDEV_ON_ERROR_STOP ||
         on_source_error == BLOCKDEV_ON_ERROR_ENOSPC) &&
        !bdrv_iostatus_is_enabled(bs)) {
        error_set(errp, QERR_INVALID_PARAMETER, "on-source-error");
        return;
    }

    if (!bdrv_is_inserted(bs)) {
        error_setg(errp, "Device is not inserted: %s",
                   bdrv_get_device_name(bs));
        return;
    }

    if (!bdrv_is_inserted(target)) {
        error_setg(errp, "Device is not inserted: %s",
                   bdrv_get_device_name(target));
        return;
    }

    if (compress && target->drv->bdrv_co_write_compressed == NULL) {
        error_setg(errp, "Compression is not supported for this drive %s",
                   bdrv_get_device_name(target));
        return;
    }

    if (bdrv_op_is_blocked(bs, BLOCK_OP_TYPE_BACKUP_SOURCE, errp)) {
        return;
    }

    if (bdrv_op_is_blocked(target, BLOCK_OP_TYPE_BACKUP_TARGET, errp)) {
        return;
    }

    if (sync_mode == MIRROR_SYNC_MODE_INCREMENTAL) {
        if (!sync_bitmap) {
            error_setg(errp, "must provide a valid bitmap name for "
                             "\"incremental\" sync mode");
            return;
        }

        /* Create a new bitmap, and freeze/disable this one. */
        if (bdrv_dirty_bitmap_create_successor(bs, sync_bitmap, errp) < 0) {
            return;
        }
    } else if (sync_bitmap) {
        error_setg(errp,
                   "a sync_bitmap was provided to backup_run, "
                   "but received an incompatible sync_mode (%s)",
                   MirrorSyncMode_lookup[sync_mode]);
        return;
    }

    len = bdrv_getlength(bs);
    if (len < 0) {
        error_setg_errno(errp, -len, "unable to get length for '%s'",
                         bdrv_get_device_name(bs));
        goto error;
    }

    BackupBlockJob *job = block_job_create(&backup_job_driver, bs, speed,
                                           cb, opaque, errp);
    if (!job) {
        goto error;
    }

    bdrv_op_block_all(target, job->common.blocker);

    job->on_source_error = on_source_error;
    job->on_target_error = on_target_error;
    job->target = target;
    job->sync_mode = sync_mode;
    job->sync_bitmap = sync_mode == MIRROR_SYNC_MODE_INCREMENTAL ?
                       sync_bitmap : NULL;
    job->compress = compress;
    job->common.len = len;
    job->common.co = qemu_coroutine_create(backup_run);
    block_job_txn_add_job(txn, &job->common);
    qemu_coroutine_enter(job->common.co, job);
    return;

 error:
    if (sync_bitmap) {
        bdrv_reclaim_dirty_bitmap(bs, sync_bitmap, NULL);
    }
}
