/*
 * Bitmaps for the QCOW version 2 format
 *
 * Copyright (c) 2014-2015 Vladimir Sementsov-Ogievskiy
 *
 * This file is derived from qcow2-snapshot.c, original copyright:
 * Copyright (c) 2004-2006 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "qemu/log.h"

#include "block/block_int.h"
#include "block/qcow2.h"

/* NOTICE: BME here means Bitmaps Extension and used as a namespace for
 * _internal_ constants. Please do not use this _internal_ abbreviation for
 * other needs and/or outside of this file. */

/* Bitmap directory entry constraints */
#define BME_MAX_TABLE_SIZE 0x8000000
#define BME_MAX_PHYS_SIZE 0x20000000 /* 512 mb */
#define BME_MAX_GRANULARITY_BITS 31
#define BME_MIN_GRANULARITY_BITS 9
#define BME_MAX_NAME_SIZE 1023

/* Bitmap directory entry flags */
#define BME_RESERVED_FLAGS 0xfffffffc
#define BME_FLAG_IN_USE 1
#define BME_FLAG_AUTO   (1U << 1)
#define BME_FLAG_EXTRA_DATA_COMPATIBLE   (1U << 1)

/* bits [1, 8] U [56, 63] are reserved */
#define BME_TABLE_ENTRY_RESERVED_MASK 0xff000000000001fe

typedef enum BitmapType {
    BT_DIRTY_TRACKING_BITMAP = 1
} BitmapType;

static BdrvDirtyBitmap *load_bitmap(BlockDriverState *bs, QCow2BitmapHeader *h,
                                    Error **errp);

void qcow2_free_bitmaps(BlockDriverState *bs)
{
    BDRVQcow2State *s = bs->opaque;

    g_free(s->bitmap_directory);
    s->bitmap_directory = NULL;
}

static void bitmap_header_to_cpu(QCow2BitmapHeader *h)
{
    be64_to_cpus(&h->bitmap_table_offset);
    be32_to_cpus(&h->bitmap_table_size);
    be32_to_cpus(&h->flags);
    be16_to_cpus(&h->name_size);
    be32_to_cpus(&h->extra_data_size);
}

static void bitmap_header_to_be(QCow2BitmapHeader *h)
{
    cpu_to_be64s(&h->bitmap_table_offset);
    cpu_to_be32s(&h->bitmap_table_size);
    cpu_to_be32s(&h->flags);
    cpu_to_be16s(&h->name_size);
    cpu_to_be32s(&h->extra_data_size);
}

static inline bool can_write(BlockDriverState *bs)
{
    return !bdrv_is_read_only(bs) && !(bdrv_get_flags(bs) & BDRV_O_INCOMING);
}

static int calc_dir_entry_size(size_t name_size)
{
    return align_offset(sizeof(QCow2BitmapHeader) + name_size, 8);
}

static int dir_entry_size(QCow2BitmapHeader *h)
{
    return calc_dir_entry_size(h->name_size);
}

static void directory_to_be(uint8_t *dir, size_t size)
{
    uint8_t *end = dir + size;
    while (dir < end) {
        QCow2BitmapHeader *h = (QCow2BitmapHeader *)dir;
        dir += dir_entry_size(h);

        bitmap_header_to_be(h);
    }
}

static int check_constraints(QCow2BitmapHeader *h, int cluster_size,
                             uint64_t disk_size)
{
    uint64_t phys_bitmap_bytes =
        (uint64_t)h->bitmap_table_size * cluster_size;
    uint64_t max_virtual_bits = (phys_bitmap_bytes * 8) << h->granularity_bits;

    int fail =
            (h->bitmap_table_offset % cluster_size) ||
            (h->bitmap_table_size > BME_MAX_TABLE_SIZE) ||
            (phys_bitmap_bytes > BME_MAX_PHYS_SIZE) ||
            (disk_size > max_virtual_bits) ||
            (h->granularity_bits > BME_MAX_GRANULARITY_BITS) ||
            (h->granularity_bits < BME_MIN_GRANULARITY_BITS) ||
            (h->flags & BME_RESERVED_FLAGS) ||
            (h->name_size > BME_MAX_NAME_SIZE) ||
            (h->type != BT_DIRTY_TRACKING_BITMAP);

    return fail ? -EINVAL : 0;
}

static inline QCow2BitmapHeader *next_dir_entry(QCow2BitmapHeader *entry)
{
    return (QCow2BitmapHeader *)((uint8_t *)entry + dir_entry_size(entry));
}

static int update_bitmap_header_sync(BlockDriverState *bs, QCow2BitmapHeader *h)
{
    int ret;
    BDRVQcow2State *s = bs->opaque;
    int sz = dir_entry_size(h);
    uint64_t off = s->bitmap_directory_offset +
                   ((uint8_t *)h - s->bitmap_directory);

    bitmap_header_to_be(h);
    ret = bdrv_pwrite(bs->file, off, h, sz);
    bitmap_header_to_cpu(h);
    if (ret < 0) {
        return ret;
    }

    ret = bdrv_flush(bs);
    if (ret < 0) {
        return ret;
    }

    return 0;
}

static int directory_read(BlockDriverState *bs, Error **errp)
{
    int ret;
    BDRVQcow2State *s = bs->opaque;
    int64_t nb_sectors = bdrv_nb_sectors(bs);
    size_t offset;
    unsigned i;

    if (nb_sectors < 0) {
        error_setg(errp, "Can't calculate number of disk sectors.");
        return nb_sectors;
    }

    if (s->bitmap_directory != NULL) {
        /* already read */
        error_setg(errp, "Try read bitmaps, when they are already read.");
        return -EEXIST;
    }

    s->bitmap_directory = g_try_malloc0(s->bitmap_directory_size);
    if (s->bitmap_directory == NULL) {
        error_setg(errp, "Can't allocate space for bitmap directory.");
        return -ENOMEM;
    }

    ret = bdrv_pread(bs->file,
                     s->bitmap_directory_offset,
                     s->bitmap_directory,
                     s->bitmap_directory_size);
    if (ret < 0) {
        error_setg(errp, "Can't read bitmap directory.");
        goto fail;
    }

    offset = 0;
    for (i = 0; i < s->nb_bitmaps; ++i) {
        QCow2BitmapHeader *h =
                (QCow2BitmapHeader *)(s->bitmap_directory + offset);

        if (offset >= s->bitmap_directory_size) {
            error_setg(errp, "Broken bitmap directory.");
            goto fail;
        }

        bitmap_header_to_cpu(h);

        ret = check_constraints(h, s->cluster_size,
                                nb_sectors << BDRV_SECTOR_BITS);
        if (ret < 0) {
            error_setg(errp, "Bitmap doesn't satisfy the constraints.");
            goto fail;
        }

        offset += dir_entry_size(h);
    }
    return 0;

fail:
    g_free(s->bitmap_directory);
    s->bitmap_directory = NULL;

    return ret;
}

static int load_autoload_bitmaps(BlockDriverState *bs, Error **errp)
{
    BDRVQcow2State *s = bs->opaque;
    QCow2BitmapHeader *h,
                      *begin = (QCow2BitmapHeader *)s->bitmap_directory,
                      *end = (QCow2BitmapHeader *)
                          (s->bitmap_directory + s->bitmap_directory_size);

    for (h = begin; h < end; h = next_dir_entry(h)) {
        uint32_t fl = h->flags;

        if ((fl & BME_FLAG_AUTO) && !(fl & BME_FLAG_IN_USE)) {
            BdrvDirtyBitmap *bitmap = load_bitmap(bs, h, errp);
            if (bitmap == NULL) {
                return -1;
            }

            bdrv_dirty_bitmap_set_internal_persistance(bitmap, true);
            bdrv_dirty_bitmap_set_autoload(bitmap, true);
        }
    }

    return 0;
}

static int bitmap_table_load(BlockDriverState *bs, QCow2BitmapHeader *bmh,
                             uint64_t **table);
int qcow2_dirty_bitmaps_refcounts(BlockDriverState *bs,
                                  BdrvCheckResult *res,
                                  void **refcount_table,
                                  int64_t *refcount_table_size)
{
    BDRVQcow2State *s = bs->opaque;
    QCow2BitmapHeader *h,
                      *begin = (QCow2BitmapHeader *)s->bitmap_directory,
                      *end = (QCow2BitmapHeader *)
                          (s->bitmap_directory + s->bitmap_directory_size);

    int i;
    int ret = inc_refcounts(bs, res, refcount_table, refcount_table_size,
                        s->bitmap_directory_offset,
                        s->bitmap_directory_size);
    if (ret < 0) {
        return ret;
    }

    for (h = begin; h < end; h = next_dir_entry(h)) {
        uint64_t *bitmap_table = NULL;

        ret = inc_refcounts(bs, res, refcount_table, refcount_table_size,
                            h->bitmap_table_offset,
                            h->bitmap_table_size);
        if (ret < 0) {
            return ret;
        }

        ret = bitmap_table_load(bs, h, &bitmap_table);
        if (ret < 0) {
            return ret;
        }

        for (i = 0; i < h->bitmap_table_size; ++i) {
            uint64_t off = be64_to_cpu(bitmap_table[i]);
            if (off <= 1) {
                continue;
            }

            ret = inc_refcounts(bs, res, refcount_table, refcount_table_size,
                                off, s->cluster_size);
            if (ret < 0) {
                g_free(bitmap_table);
                return ret;
            }
        }

        g_free(bitmap_table);
    }

    return 0;
}

static int directory_del_in_use(BlockDriverState *bs);
int qcow2_read_bitmaps(BlockDriverState *bs, Error **errp)
{
    int ret;
    BDRVQcow2State *s = bs->opaque;

    if (s->bitmap_directory != NULL) {
        /* already read */
        error_setg(errp, "Try read bitmaps, when they are already read.");
        return -EEXIST;
    }

    if (s->nb_bitmaps == 0) {
        /* No bitmaps - nothing to do */
        return 0;
    }

    ret = directory_read(bs, errp);
    if (ret < 0) {
        return ret;
    }

    if (can_write(bs)) {
        ret = directory_del_in_use(bs);
        if (ret < 0) {
            error_setg_errno(errp, -ret, "Can't delete in_use bitmaps.");
            goto fail;
        }

        if (s->nb_bitmaps == 0) {
            /* No bitmaps - nothing to do */
            return 0;
        }
    }

    ret = load_autoload_bitmaps(bs, errp);
    if (ret < 0) {
        goto fail;
    }

    return 0;

fail:
    qcow2_free_bitmaps(bs);

    return ret;
}

static QCow2BitmapHeader *find_bitmap_by_name(BlockDriverState *bs, const char *name)
{
    BDRVQcow2State *s = bs->opaque;
    QCow2BitmapHeader *h,
                      *begin = (QCow2BitmapHeader *)s->bitmap_directory,
                      *end = (QCow2BitmapHeader *)
                          (s->bitmap_directory + s->bitmap_directory_size);

    for (h = begin; h < end; h = next_dir_entry(h)) {
        if (strncmp((char *)(h+1), name, h->name_size) == 0) {
            return h;
        }
    }

    return NULL;
}

/* load_bitmap_data()
 * load dirty bitmap from bitmap table
 * Bitmap table entries are assumed to be in big endian format */
static int load_bitmap_data(BlockDriverState *bs, const uint64_t *bitmap_table,
                            uint32_t bitmap_table_size, BdrvDirtyBitmap *bitmap)
{
    int ret = 0;
    BDRVQcow2State *s = bs->opaque;
    uint32_t i;
    uint64_t *tab = g_memdup(bitmap_table,
                             bitmap_table_size * sizeof(uint64_t));

    for (i = 0; i < bitmap_table_size; ++i) {
        be64_to_cpus(tab + i);
    }

    ret = bdrv_dirty_bitmap_load(bitmap, bs->file, tab, bitmap_table_size,
                                 s->cluster_size);

    g_free(tab);

    return ret;
}

bool qcow2_bitmap_load_check(BlockDriverState *file, const char *name)
{
    return find_bitmap_by_name(file, name) != NULL;
}

static int validate_bitmap_table(BlockDriverState *file,
                                 uint64_t *bitmap_table, size_t size)
{
    size_t i;
    int64_t fsize = bdrv_nb_sectors(file);

    if (fsize < 0) {
        return fsize;
    }

    fsize <<= BDRV_SECTOR_BITS;

    for (i = 0; i < size; ++i) {
        assert(be64_to_cpu(bitmap_table[i]) < fsize);
    }

    return 0;
}

static BdrvDirtyBitmap *load_bitmap(BlockDriverState *bs,
                                    QCow2BitmapHeader *bmh, Error **errp)
{
    int ret;
    uint64_t *bitmap_table = NULL;
    uint32_t granularity;
    BdrvDirtyBitmap *bitmap = NULL;
    char *name = g_strndup((char *)(bmh + 1), bmh->name_size);

    if (bmh->flags & BME_FLAG_IN_USE) {
        error_setg(errp, "Bitmap '%s' is in use", name);
        goto fail;
    }

    if (!(bmh->flags & BME_FLAG_EXTRA_DATA_COMPATIBLE) &&
            bmh->extra_data_size != 0) {
        error_setg(errp, "Uncompatible extra data found for bitmap '%s'",
                   name);
        goto fail;
    }

    ret = bitmap_table_load(bs, bmh, &bitmap_table);
    if (ret < 0) {
        error_setg_errno(errp, -ret,
                         "Could not read bitmap_table table from image");
        goto fail;
    }

    granularity = 1U << bmh->granularity_bits;

    bitmap = bdrv_find_dirty_bitmap(bs, name);
    if (bitmap == NULL) {
        bitmap = bdrv_create_dirty_bitmap(bs, granularity, name, errp);
        if (bitmap == NULL) {
            goto fail;
        }

        ret = load_bitmap_data(bs, bitmap_table, bmh->bitmap_table_size,
                               bitmap);
        if (ret < 0) {
            error_setg_errno(errp, -ret, "Could not read bitmap from image");
            goto fail;
        }
    } else if (granularity != bdrv_dirty_bitmap_granularity(bitmap)) {
        error_setg(errp, "Bitmap '%s' already exists with other granularity",
                   name);
        goto fail;
    }

    if (can_write(bs)) {
        bmh->flags |= BME_FLAG_IN_USE;
        ret = update_bitmap_header_sync(bs, bmh);
        if (ret < 0) {
            error_setg_errno(errp, -ret,
                             "Could not set in_use in bitmap header");
            goto fail;
        }
    }

    g_free(name);
    g_free(bitmap_table);
    return bitmap;

fail:
    g_free(name);
    g_free(bitmap_table);
    if (bitmap != NULL) {
        bdrv_release_dirty_bitmap(bs, bitmap);
    }

    return NULL;
}

BdrvDirtyBitmap *qcow2_bitmap_load(BlockDriverState *bs, const char *name,
                                   Error **errp)
{
    QCow2BitmapHeader *h = find_bitmap_by_name(bs, name);
    if (h == NULL) {
        error_setg(errp, "Could not find bitmap '%s' in the node '%s'", name,
                   bdrv_get_device_or_node_name(bs));
        return NULL;
    }

    return load_bitmap(bs, h, errp);
}

static int update_header_sync(BlockDriverState *bs)
{
    int ret;

    ret = qcow2_update_header(bs);
    if (ret < 0) {
        return ret;
    }

    ret = bdrv_flush(bs);
    if (ret < 0) {
        return ret;
    }

    return 0;
}

/* write bitmap directory from the state to new allocated clusters */
static int64_t directory_write(BlockDriverState *bs, const uint8_t *dir,
                               size_t size)
{
    int ret = 0;
    uint8_t *dir_be = NULL;
    int64_t dir_offset = 0;

    dir_be = g_try_malloc(size);
    if (dir_be == NULL) {
        return -ENOMEM;
    }
    memcpy(dir_be, dir, size);
    directory_to_be(dir_be, size);

    /* Allocate space for the new bitmap directory */
    dir_offset = qcow2_alloc_clusters(bs, size);
    if (dir_offset < 0) {
        ret = dir_offset;
        goto out;
    }

    /* The bitmap directory position has not yet been updated, so these
     * clusters must indeed be completely free */
    ret = qcow2_pre_write_overlap_check(bs, 0, dir_offset, size);
    if (ret < 0) {
        goto out;
    }

    ret = bdrv_pwrite(bs->file, dir_offset, dir_be, size);
    if (ret < 0) {
        goto out;
    }

out:
    g_free(dir_be);

    if (ret < 0) {
        if (dir_offset > 0) {
            qcow2_free_clusters(bs, dir_offset, size, QCOW2_DISCARD_ALWAYS);
        }

        return ret;
    }

    return dir_offset;
}

static int directory_update(BlockDriverState *bs, uint8_t *new_dir,
                            size_t new_size, uint32_t new_nb_bitmaps)
{
    BDRVQcow2State *s = bs->opaque;
    int ret;
    int64_t new_offset = 0;
    uint64_t old_offset = s->bitmap_directory_offset;
    uint64_t old_size = s->bitmap_directory_size;
    uint64_t old_autocl = s->autoclear_features;
    uint32_t old_nb_bitmaps = s->nb_bitmaps;

    if (new_size > QCOW_MAX_DIRTY_BITMAP_DIRECTORY_SIZE) {
        return -EINVAL;
    }

    if ((new_size == 0) != (new_nb_bitmaps == 0)) {
        return -EINVAL;
    }

    if (new_nb_bitmaps > 0) {
        new_offset = directory_write(bs, new_dir, new_size);
        if (new_offset < 0) {
            return new_offset;
        }

        ret = bdrv_flush(bs);
        if (ret < 0) {
            goto fail;
        }

        s->autoclear_features |= QCOW2_AUTOCLEAR_DIRTY_BITMAPS;
    } else {
        s->autoclear_features &= ~(uint64_t)QCOW2_AUTOCLEAR_DIRTY_BITMAPS;
    }
    s->bitmap_directory_offset = new_offset;
    s->bitmap_directory_size = new_size;
    s->nb_bitmaps = new_nb_bitmaps;

    ret = update_header_sync(bs);
    if (ret < 0) {
        goto fail;
    }

    if (old_size) {
        qcow2_free_clusters(bs, old_offset, old_size, QCOW2_DISCARD_ALWAYS);
    }

    g_free(s->bitmap_directory);
    s->bitmap_directory = new_dir;

    return 0;

fail:
    if (new_offset > 0) {
        qcow2_free_clusters(bs, new_offset, new_size, QCOW2_DISCARD_ALWAYS);
        s->bitmap_directory_offset = old_offset;
        s->bitmap_directory_size = old_size;
        s->autoclear_features = old_autocl;
        s->nb_bitmaps = old_nb_bitmaps;
    }

    return ret;
}

static void clear_bitmap_table(BlockDriverState *bs, uint64_t *bitmap_table,
                               uint32_t bitmap_table_size)
{
    BDRVQcow2State *s = bs->opaque;
    int cl_size = s->cluster_size;
    int i;

    for (i = 0; i < bitmap_table_size; ++i) {
        uint64_t addr = be64_to_cpu(bitmap_table[i]);
        if (addr <= 1) {
            continue;
        }

        qcow2_free_clusters(bs, addr, cl_size, QCOW2_DISCARD_ALWAYS);
        bitmap_table[i] = 0;
    }
}

static int free_bitmap_clusters(BlockDriverState *bs, QCow2BitmapHeader *bmh)
{
    int ret;
    uint64_t *bitmap_table;

    ret = bitmap_table_load(bs, bmh, &bitmap_table);
    if (ret < 0) {
        return ret;
    }

    clear_bitmap_table(bs, bitmap_table, bmh->bitmap_table_size);

    qcow2_free_clusters(bs, bmh->bitmap_table_offset, bmh->bitmap_table_size *
                        sizeof(uint64_t), QCOW2_DISCARD_ALWAYS);

    return 0;
}

int qcow2_delete_bitmaps(BlockDriverState *bs)
{
    BDRVQcow2State *s = bs->opaque;
    QCow2BitmapHeader *h,
                      *begin = (QCow2BitmapHeader *)s->bitmap_directory,
                      *end = (QCow2BitmapHeader *)
                          (s->bitmap_directory + s->bitmap_directory_size);

    if (s->nb_bitmaps == 0) {
        return 0;
    }

    for (h = begin; h < end; h = next_dir_entry(h)) {
        free_bitmap_clusters(bs, h);
    }

    return directory_update(bs, NULL, 0, 0);
}

static int directory_del_in_use(BlockDriverState *bs)
{
    int ret;
    BDRVQcow2State *s = bs->opaque;
    uint64_t new_size = 0;
    uint32_t new_nb_bitmaps = 0;
    uint8_t *new_dir = NULL, *p;
    QCow2BitmapHeader *h,
                      *begin = (QCow2BitmapHeader *)s->bitmap_directory,
                      *end = (QCow2BitmapHeader *)
                          (s->bitmap_directory + s->bitmap_directory_size);

    for (h = begin; h < end; h = next_dir_entry(h)) {
        if (!(h->flags & BME_FLAG_IN_USE)) {
            new_size += dir_entry_size(h);
            new_nb_bitmaps++;
        }
    }

    if (new_nb_bitmaps == s->nb_bitmaps) {
        return 0;
    }

    /* free in_use bitmaps clusters */
    for (h = begin; h < end; h = next_dir_entry(h)) {
        if (h->flags & BME_FLAG_IN_USE) {
            char *bitmap_name = g_strndup((char *)(h + 1), h->name_size);
            qemu_log("Delete corrupted (in_use) bitmap '%s' from device '%s'\n",
                     bitmap_name, bdrv_get_device_or_node_name(bs));
            g_free(bitmap_name);

            free_bitmap_clusters(bs, h);
        }
    }

    if (new_nb_bitmaps == 0) {
        goto update;
    }

    p = new_dir = g_try_malloc(new_size);
    if (new_dir == NULL) {
        return -ENOMEM;
    }

    for (h = begin; h < end; ) {
        QCow2BitmapHeader *h2;
        size_t copy_size;

        for (; h < end && (h->flags & BME_FLAG_IN_USE); h = next_dir_entry(h)) {
            ;
        }

        if (h >= end) {
            break;
        }

        for (h2 = next_dir_entry(h); h2 < end && !(h->flags & BME_FLAG_IN_USE);
                h2 = next_dir_entry(h)) {
            ;
        }

        copy_size = (uint8_t *)h2 - (uint8_t *)h;
        memcpy(p, h, copy_size);
        p += copy_size;
        h = h2;
    }

update:
    ret = directory_update(bs, new_dir, new_size, new_nb_bitmaps);
    if (ret < 0) {
        g_free(new_dir);
        return ret;
    }

    return 0;
}


static int directory_push_entry(BlockDriverState *bs, QCow2BitmapHeader *header)
{
    BDRVQcow2State *s = bs->opaque;
    int ret;
    int entry_size = dir_entry_size(header);
    uint64_t new_size = s->bitmap_directory_size + entry_size;
    int64_t nb_sectors = bdrv_nb_sectors(bs);

    if (nb_sectors < 0) {
        return nb_sectors;
    }

    if (new_size > QCOW_MAX_DIRTY_BITMAP_DIRECTORY_SIZE) {
        return -EINVAL;
    }

    ret = check_constraints(header, s->cluster_size,
                            nb_sectors << BDRV_SECTOR_BITS);
    if (ret < 0) {
        return -EINVAL;
    }

    uint8_t *new_dir = g_try_malloc(new_size);
    if (new_dir == NULL) {
        return -ENOMEM;
    }
    memcpy(new_dir, s->bitmap_directory, s->bitmap_directory_size);
    memcpy(new_dir + s->bitmap_directory_size, header, entry_size);

    ret = directory_update(bs, new_dir, new_size, s->nb_bitmaps + 1);
    if (ret < 0) {
        g_free(new_dir);
        return ret;
    }

    return 0;
}

/* store_bitmap()
 * Store bitmap to qcow2 and set bitmap_table. bitmap_table itself is not
 * stored to qcow2.
 */
static int store_bitmap(BlockDriverState *bs, uint64_t *bitmap_table,
                        uint32_t bitmap_table_size,
                        const BdrvDirtyBitmap *bitmap)
{
    int ret;
    BDRVQcow2State *s = bs->opaque;
    int cl_size = s->cluster_size;
    uint32_t i, tab_size = 0;

    bdrv_dirty_bitmap_prepare_store(bitmap, s->cluster_size, NULL, &tab_size);
    if (tab_size != bitmap_table_size) {
        return -EINVAL;
    }

    ret = bdrv_dirty_bitmap_prepare_store(bitmap, s->cluster_size,
                                          bitmap_table, &tab_size);
    if (ret < 0) {
        return ret;
    }

    for (i = 0; i < bitmap_table_size; ++i) {
        int64_t off;

        if (bitmap_table[i] <= 1) {
            continue;
        }

        off = qcow2_alloc_clusters(bs, cl_size);
        if (off < 0) {
            ret = off;
            goto fail;
        }

        bitmap_table[i] = off;
    }

    ret = bdrv_dirty_bitmap_store(bitmap, bs->file, bitmap_table,
                                  bitmap_table_size, cl_size);
    if (ret < 0) {
        goto fail;
    }

    for (i = 0; i < bitmap_table_size; ++i) {
        cpu_to_be64s(&bitmap_table[i]);
    }

    return 0;

fail:
    for (i = 0 ; i < bitmap_table_size; ++i) {
        if (bitmap_table[i] <= 1) {
            continue;
        }

        qcow2_free_clusters(bs, bitmap_table[i], cl_size, QCOW2_DISCARD_ALWAYS);
    }
    memset(bitmap_table, 0, bitmap_table_size * sizeof(bitmap_table[0]));

    return ret;
}

static int64_t alloc_zeroed_clusters(BlockDriverState *bs, uint64_t size)
{
    int ret = 0;
    void *buf = NULL;
    int64_t offset = qcow2_alloc_clusters(bs, size);
    if (offset < 0) {
        return offset;
    }

    buf = g_try_malloc0(size);
    if (buf == NULL) {
        ret = -ENOMEM;
        goto out;
    }

    ret = qcow2_pre_write_overlap_check(bs, 0, offset, size);
    if (ret < 0) {
        goto out;
    }

    ret = bdrv_pwrite(bs->file, offset, buf, size);
    if (ret < 0) {
        goto out;
    }

    ret = bdrv_flush(bs);
    if (ret < 0) {
        goto out;
    }

out:
    g_free(buf);

    if (ret < 0) {
        qcow2_free_clusters(bs, offset, size, QCOW2_DISCARD_ALWAYS);
        return ret;
    }

    return offset;
}

static int directory_push(BlockDriverState *bs, const char *name,
                          uint64_t size, int granularity)
{
    int ret;
    BDRVQcow2State *s = bs->opaque;
    int sector_granularity = granularity >> BDRV_SECTOR_BITS;
    size_t name_size = strlen(name);
    size_t entry_size = calc_dir_entry_size(name_size);
    QCow2BitmapHeader *entry = g_malloc0(entry_size);
    int64_t table_offset = 0;

    entry->granularity_bits = ctz32(granularity);
    entry->type = BT_DIRTY_TRACKING_BITMAP;
    entry->name_size = name_size;
    memcpy(entry + 1, name, name_size);

    entry->bitmap_table_size =
            size_to_clusters(s, (((size - 1) / sector_granularity) >> 3) + 1);
    table_offset = alloc_zeroed_clusters(bs, entry->bitmap_table_size *
                                  sizeof(uint64_t));
    if (table_offset < 0) {
        ret = table_offset;
        goto out;
    }
    entry->bitmap_table_offset = table_offset;

    ret = directory_push_entry(bs, entry);
    if (ret < 0) {
        goto out;
    }

out:
    g_free(entry);
    if (ret < 0 && table_offset > 0) {
        qcow2_free_clusters(bs, table_offset, entry->bitmap_table_size *
                            sizeof(uint64_t), QCOW2_DISCARD_ALWAYS);
    }

    return ret;
}

/* if no id is provided, a new one is constructed */
static int qcow2_bitmap_create(BlockDriverState *bs, const char *name,
                               uint64_t size, int granularity)
{
    BDRVQcow2State *s = bs->opaque;

    if (s->nb_bitmaps >= QCOW_MAX_DIRTY_BITMAPS) {
        return -EFBIG;
    }

    /* Check that the name is unique */
    if (find_bitmap_by_name(bs, name) != NULL) {
        return -EEXIST;
    }

    return directory_push(bs, name, size, granularity);
}

static int bitmap_table_load(BlockDriverState *bs, QCow2BitmapHeader *bmh,
                             uint64_t **table)
{
    int ret;

    *table = g_try_new(uint64_t, bmh->bitmap_table_size);
    if (*table == NULL) {
        return -ENOMEM;
    }

    ret = bdrv_pread(bs->file, bmh->bitmap_table_offset,
                     *table, bmh->bitmap_table_size * sizeof(uint64_t));
    if (ret < 0) {
        g_free(*table);
        *table = NULL;
        return ret;
    }

    return validate_bitmap_table(bs->file, *table, bmh->bitmap_table_size);
}

void qcow2_bitmap_store(BlockDriverState *bs,
                        const BdrvDirtyBitmap *bitmap, Error **errp)
{
    int ret = 0;
    uint64_t *bitmap_table;
    QCow2BitmapHeader *bmh;
    const char *name = bdrv_dirty_bitmap_name(bitmap);
    uint64_t size = bdrv_dirty_bitmap_size(bitmap);
    int granularity = bdrv_dirty_bitmap_granularity(bitmap);

    if (!bdrv_has_dirty_bitmap(bs, bitmap)) {
        error_setg(errp, "Can't store bitmap to the other node.");
        return;
    }

    /* find/create dirty bitmap */
    bmh = find_bitmap_by_name(bs, name);
    if (bmh == NULL) {
        ret = qcow2_bitmap_create(bs, name, size, granularity);
        if (ret < 0) {
            error_setg_errno(errp, ret, "Can't create dirty bitmap in qcow2.");
            return;
        }
        bmh = find_bitmap_by_name(bs, name);
        if (bmh == NULL) {
            error_setg(errp, "Can't create dirty bitmap in qcow2.");
            return;
        }
    } else {

        if (granularity != (1U << bmh->granularity_bits)) {
            error_setg(errp,
                       "The bitmap with same name (but other granularity) "
                       "already exists.");
            return;
        }
    }

    ret = bitmap_table_load(bs, bmh, &bitmap_table);
    if (ret < 0) {
        error_setg_errno(errp, ret, "Can't load dirty bitmap table.");
        return;
    }

    clear_bitmap_table(bs, bitmap_table, bmh->bitmap_table_size);

    ret = store_bitmap(bs, bitmap_table, bmh->bitmap_table_size,
                       bitmap);
    if (ret < 0) {
        error_setg_errno(errp, ret, "Can't store bitmap table.");
        goto finish;
    }

    ret = validate_bitmap_table(bs->file,
                                bitmap_table, bmh->bitmap_table_size);
    if (ret < 0) {
        error_setg_errno(errp, ret, "Can't validate dirty bitmap table.");
        goto finish;
    }
    ret = bdrv_pwrite(bs->file, bmh->bitmap_table_offset,
                      bitmap_table,
                      bmh->bitmap_table_size * sizeof(uint64_t));
    if (ret < 0) {
        error_setg_errno(errp, ret, "Can't write dirty bitmap table.");
        goto finish;
    }

    bmh->flags &= ~BME_FLAG_IN_USE;
    bmh->flags &= ~BME_FLAG_AUTO;
    if (bdrv_dirty_bitmap_get_autoload(bitmap)) {
        bmh->flags |= BME_FLAG_AUTO;
    }

    ret = update_bitmap_header_sync(bs, bmh);
    if (ret < 0) {
        error_setg_errno(errp, ret, "Can't update bitmap header.");
        goto finish;
    }

finish:
    g_free(bitmap_table);
}
